// ----------------------------------------------------------------------------
// Assume.java
// The "assume" command.
// $Id: Assume.java,v 1.10 2006/03/29 15:49:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.Proving.*;
import fmrisc.ProofNavigator.*;

/**
 * 
 * Commnand "assume": split proof state by an assumption and its proof.
 *
 */
public final class Assume extends ProofCommandBase
{
  public final static String Name = "assume";
  private Expression formula; // the formula splitting the state
  
  /***************************************************************************
   * Create an "assume" command.
   * @param formula the formula assumed in one state and proved in the other.
   **************************************************************************/
  public Assume(Expression formula)
  {
    super(Name);
    this.formula = formula;
  }
 
  /****************************************************************************
   * Process command "assume": split current proof state by formula F.
   * A, ... |- B, ... =>
   *  (A, ... |- TCC_F     TCC_F type checking condition for F)
   *   A, ..., F |- B, ... 
   *   A, ..., ~B, ~... |- F
   * @return null or array of two or three new proof states
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Environment env = state.getEnvironment();
    Context context = State.getContext();
    context.open();
    if (!typeCheckFormula(env, formula)) 
    {
      context.abort();
      return null;
    }
    Expression tcc = context.getFormula();
    context.abort();
    
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    
    ProofState state0 = new ProofState(state, env);
    Formula[] assumptions0 = state0.add(assumptions,
      new Formula(state0, formula, false));
    Formula[] goals0 = state0.copy(goals);
    if (tcc != null)
      state0.setAxioms(state0.add(state0.getAxioms(),
        new Formula(state0, tcc, false)));
    state0.setAssumptions(assumptions0);
    state0.setGoals(goals0);
    
    ProofState state1 = new ProofState(state, env);
    Vector assumptions1 = new Vector();
    int an = assumptions.length;
    for (int i=0; i<an; i++)
    {
      Formula f = assumptions[i];
      assumptions1.add(new Formula(state1, f.getExpression(), false));
    }
    int gn = goals.length;
    for (int i=0; i<gn; i++)
    {
      Formula f = goals[i];
      assumptions1.add(new Formula(state1, 
          ProofUtil.negate(f.getExpression()), false));
    }
    
    Formula[] goals1 = { new Formula(state1, formula, true) };
    if (tcc != null)
      state1.setAxioms(state1.add(state1.getAxioms(),
        new Formula(state1, tcc, false)));
    state1.setAssumptions(Formula.toArray(assumptions1));
    state1.setGoals(goals1);
    
    if (tcc == null) return new ProofState[] { state0, state1 };
    
    ProofState state2 = new ProofState(state, env);
    Formula[] assumptions2 = state2.copy(assumptions);
    Formula[] goals2 = { new Formula(state2, tcc, true) };
    state2.setAssumptions(assumptions2);
    state2.setGoals(goals2);
    
    return new ProofState[] { state2, state0, state1 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(toNode(formula));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    if (children.getLength() != 1) 
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    Expression formula = toExpression(children.item(0));
    if (formula == null) return null;
    return new Assume(formula);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    formula.print(out);
  }
}
