// ----------------------------------------------------------------------------
// Lemma.java
// The "lemma" command.
// $Id: Lemma.java,v 1.9 2006/03/29 15:49:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.PrintWriter;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.Proving.*;
import fmrisc.ProofNavigator.*;

/**
 * 
 * The "lemma" command: import other formulas.
 *
 */
public final class Lemma extends ProofCommandBase
{
  public final static String Name = "lemma";
  private Identifier[] idents ;    // the names of the formulas to be imported
  
  /***************************************************************************
   * Create a "lemma" command.
   * @param idents the identifiers of the formulas to be imported.
   **************************************************************************/
  public Lemma(Identifier[] idents)
  {
    super(Name);
    this.idents = idents;
  }
  
  /***************************************************************************
   * Return symbols of formulas referenced in lemma.
   * @param env the environment in which the symbols are looked up.
   * @return the sequence of formula symbols.
   **************************************************************************/
  public FormulaSymbol[] getSymbols(Environment env)
  {
    int n = idents.length;
    FormulaSymbol[] symbols = new FormulaSymbol[n];
    for (int i=0; i<n; i++)
    {
      symbols[i] = env.getFormulaSymbol(idents[i]);
      if (symbols[i] == null)
      {
        error("unknown formula " + idents[i].toString());
        return null;
      }
    }
    return symbols;
  }
  
  /****************************************************************************
   * Process command "lemma": import declared lemmas as assumptions.
   * A, ... |- B => A, ..., L |- B
   * @return null or array of a single proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Environment env = state.getEnvironment();
    FormulaSymbol[] symbols = getSymbols(env);
    if (symbols == null) return null;
    ProofState state0 = new ProofState(state, env);
    int n = idents.length;
    Formula[] formulas = new Formula[n];
    for (int i=0; i<n; i++)
    {
      // replace identifier by symbol identifier for
      // later collection of symbols referenced in proof
      idents[i] = Construct.reference(symbols[i]);
      formulas[i] = new Formula(state0, symbols[i].getFormula(), false);
    }
    Formula[] assumptions0 = state0.add(state.getAssumptions(), formulas);
    Formula[] goals0 = state0.copy(state.getGoals());
    state0.setAssumptions(assumptions0);
    state0.setGoals(goals0);
    return new ProofState[] { state0 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    int n = idents.length;
    Proof proof = State.getProof();
    if (proof != null) 
    {
      for (int i=0; i<n; i++)
        proof.referenced(idents[i]);
    }
    for (int i=0; i<n; i++)
      node.appendChild(toNode(idents[i]));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0) 
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    Identifier[] idents = new Identifier[n];
    for (int i=0; i<n; i++)
    {
      idents[i] = toIdentifier(children.item(i));
      if (idents[i] == null) return null;
    }
    return new Lemma(idents);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    int n = idents.length;
    for (int i=0; i<n; i++)
    {
      idents[i].print(out);
      if (i+1 < n) out.print(", ");
    }
  }
}
