// ----------------------------------------------------------------------------
// Prove.java
// The "prove" command.
// $Id: Prove.java,v 1.14 2007/10/23 20:09:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import fmrisc.ProofNavigator.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.Proving.*;

/**
 * 
 * "prove": prove goal formula.
 *
 */
public final class Prove extends CommandBase
{
  public final static String Name = "prove";
  private Identifier ident; // reference to the formula to be proved
  
  /***************************************************************************
   * create a prove command
   * @param ident name of goal formula to be proved
   **************************************************************************/
  public Prove(Identifier ident)
  {
    super(Name);
    this.ident = ident;
  }
  
  /***************************************************************************
   * get name of formula to be proved
   * @return name of formula to be proved
   **************************************************************************/
  public Identifier getIdent()
  {
    return ident;
  }
  
  /****************************************************************************
   * process command "prove f": start proof of formula denoted by identifier f.
   ***************************************************************************/
  public void process()
  {
    PrintWriter out = State.getSystemOut();
    Proof proof = State.getProof();
    if (proof != null)
    {
      error("cannot start proof within another proof");
      return;
    }
    Environment env = State.getEnvironment();
    FormulaSymbol symbol = env.getFormulaSymbol(ident);
    if (symbol == null)
    {
      error("there is no formula called " + ident.toString());
      return;
    }
    if (symbol.isAxiom())
    {
      error("axiom " + ident.toString() + " does not need a proof");
      return;
    }
    proof = symbol.getProof();
    if (proof != null)
    {
      out.print("Formula " + ident.toString() + " already has a ");
      if (proof.isSkeleton())
        out.print("(skeleton) ");
      out.print("proof (");
      proof.printStatus(out);
      out.println(")");
      if (proof.isSkeleton())
      {
        boolean yes = Main.ask("Replay skeleton proof");
        if (yes) 
        {
          if (!proof.isTrusted())
            out.println("Warning: proof is untrusted, errors may occur.");
          boolean okay = proof.replay();
          if (okay)
            out.println("Proof replay successful.");
          else
            out.println("Proof replay had errors.");
        }
        else 
          proof = null;
      }
      else 
      {
        boolean yes = Main.ask("Open existing proof");
        if (!yes) proof = null;
      }
    }
    if (proof == null)
    {
      out.println("Proof of formula " + ident.toString() + ".");
      proof = Proof.newProof(symbol, State.useAutoSimplify());
      if (proof == null) return;
      State.setProof(proof);
    }
    else if (proof.isClosed())
    {
      out.println("Use 'proof " + ident.toString() + "' to see proof.");
      boolean okay = State.getPresenter().updateDeclarations();
      if (!okay) 
      {
        error("could not update declarations");
        return;
      }
      Main.displayDeclarations();
    }
    else
      State.setProof(proof);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    ident.print(out);
  }
}
