// ----------------------------------------------------------------------------
// Split.java
// The "split" command
// $Id: Split.java,v 1.14 2006/03/29 15:49:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.ProofNavigator.State;
import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

import java.io.PrintWriter;
import java.util.*;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * 
 * Command "split": split a proof state.
 *
 */
public final class Split extends ProofCommandBase
{
  public final static String Name = "split";
  private String label; // the formula label triggering the split
  
  /***************************************************************************
   * Create a "split" command.
   * @param label the label of the formula triggering the split
   *        (null, if the single goal is intended to trigger the split)
   **************************************************************************/
  public Split(String label)
  {
    super(Name);
    this.label = label;
  }
  
  /****************************************************************************
   * Process command "split". 
   * Split the current proof state by the following rules:
   * - A1 \/ ... \/ An, ... |- B, ... => 
   *    (A1, ... |- B, ...) ... (An, ... |- B, ...)
   * - A, ... |- B1 /\ ... /\ Bm, ... =>
   *     (A, ... |- B1, ...) ... (A, ... |- Bm, ...)
   * - A1 => A2, ... |- B, ... => 
   *    (~A1, ... |- B, ...) , (A2, ... |- B, ...)   
   * - A, ... |- B1 <=> B2, ... =>
   *     (A, ... |- B1 => B2, ...) , (A, ... |- B2 => B1, ...) 
   * - A, ... |- B1 xor B2, ... =>
   *     (A, ... |- B1 => ~B2, ...) , (A, ... |- ~B1 => B2, ...)
   * @return null or array of new proof states
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Formula sformula;
    String label0 = label;
    if (label0 == null)
    {
      Formula[] goals = state.getGoals();
      if (goals.length != 1)
      {
        error("there must be exactly one goal to be split");
        return null;
      }
      sformula = goals[0];
      label0 = sformula.getLabel();
    }
    else
      sformula = state.getFormula(label0);
    Expression[] formulas = null;
    if (sformula == null)
    {
      error("No formula with label " + label0);
      return null;
    }
    if (sformula.isGoal())
    {
      Expression exp = sformula.getExpression();
      if (exp instanceof AndFormula || exp instanceof EquivalentFormula ||
          exp instanceof NotEquivalentFormula)
        formulas = split(exp, true);
      else
      {
        error("goal " + label0 + " must be a conjunction or (non)equivalence");
        return null;
      }
    }
    else
    {
      Expression exp = sformula.getExpression();
      if (exp instanceof OrFormula || exp instanceof ImpliesFormula)
        formulas = split(exp, false);
      else
      {
        error("assumption " + label0 + " must be a disjunction or implication");
        return null;
      }
    }
    Environment env = state.getEnvironment();
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    int n = formulas.length;
    ProofState[] children = new ProofState[n];
    if (sformula.isGoal()) 
    {
      for (int i=0; i<n; i++)
      {
        ProofState cstate = new ProofState(state, env);
        children[i] = cstate;
        cstate.setAssumptions(cstate.copy(assumptions));
        cstate.setGoals(cstate.replace(goals, sformula, 
          new Formula(cstate, formulas[i], true)));
      }
    }
    else
    {
      for (int i=0; i<n; i++)
      {
        ProofState cstate = new ProofState(state, env);
        children[i] = cstate;
        cstate.setGoals(cstate.copy(goals));
        cstate.setAssumptions(cstate.replace(assumptions, sformula, 
          new Formula(cstate, formulas[i], false)));
      }
    }
    return children;
  }
  
  /****************************************************************************
   * Split a (nested) conjunction or disjunction into its components.
   * @param formula a formula to be split (remains unchanged).
   * @param conjunction if true, split conjunctions, else disjunctions
   * @return sequence of top-level conjuncts or disjuncts of formula
   **************************************************************************/
  private Expression[] split(Expression formula, boolean conjunction)
  {
    Vector vector = new Vector();
    split(formula, conjunction, vector);
    return ASTUtil.toExpressionArray(vector);
  }
  
  /****************************************************************************
   * Split a (nested) conjunction or disjunction into its components.
   * Also works on equivalences (treated as conjunction of two implications)
   * and implications (treated as disjunction of negated hypothesis 
   * and conclusion).
   * @param formula a formula to be split (remains unchanged).
   * @param conjunction if true, split conjunctions, else disjunctions
   * @param vector vector to add the component to
   **************************************************************************/
  private void split(Expression formula, boolean conjunction, Vector vector)
  {
    Expression first;
    Expression second;
    if (conjunction)
    {
      if (formula instanceof EquivalentFormula)
      {
        EquivalentFormula f = (EquivalentFormula)formula;
        first = Construct.impliesFormula(f.getFirst(), f.getSecond());
        second = Construct.impliesFormula(f.getSecond(), f.getFirst());      
      }
      else if (formula instanceof NotEquivalentFormula)
      {
        NotEquivalentFormula f = (NotEquivalentFormula)formula;
        first = Construct.impliesFormula(f.getFirst(), 
          Construct.notFormula(f.getSecond()));
        second = Construct.impliesFormula(Construct.notFormula(f.getFirst()), 
          f.getSecond());     
      }
      else if (formula instanceof AndFormula)
      {
        AndFormula f = (AndFormula)formula;
        first = f.getFirst();
        second = f.getSecond();
      }
      else
        return;
    }
    else
    {
      if (formula instanceof ImpliesFormula)
      {
        ImpliesFormula f = (ImpliesFormula)formula;
        first = Construct.notFormula(f.getFirst());
        second = f.getSecond();
      }
      else if (formula instanceof OrFormula)
      {
        OrFormula f = (OrFormula)formula;
        first = f.getFirst();
        second = f.getSecond();
      }
      else
        return;
    }
    if ((conjunction && 
          (first instanceof AndFormula || first instanceof EquivalentFormula
            || first instanceof NotEquivalentFormula)) ||
        (!conjunction && 
          (first instanceof OrFormula || first instanceof ImpliesFormula))) 
      split(first, conjunction, vector);
    else
      vector.add(first);
    if ((conjunction && 
          (second instanceof AndFormula || second instanceof EquivalentFormula
            || second instanceof NotEquivalentFormula)) ||
        (!conjunction && 
          (second instanceof OrFormula || second instanceof ImpliesFormula))) 
      split(second, conjunction, vector);
    else
      vector.add(second); 
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    if (label != null)
      node.appendChild(newTextNode(label));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n > 1) 
    {
      State.setError("at most one child expected: " + node.toString());
      return null;
    }
    if (n == 0) return new Split(null);
    String label = toString(children.item(0));
    if (label == null) return null;
    return new Split(label);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (label == null) return;
    out.print(" " + label);
  }
}
