// ----------------------------------------------------------------------------
// Undo.java
// The "undo" command.
// $Id: Undo.java,v 1.6 2007/10/23 20:09:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;

import fmrisc.Proving.*;

/**
 * 
 * The "undo" command: goto a closed proof state and undo its proof.
 *
 */
public final class Undo extends ProofCommandBase
{
  public final static String Name = "undo";
  private String label; // the label of the proof state to go to (may be null)
  
  /***************************************************************************
   * create a "undo" command
   * @param label the label of the proof state to go to (may be null)
   **************************************************************************/
  public Undo(String label)
  {
    super(Name);
    this.label = label;
  }
  
  /****************************************************************************
   * Process command "undo": switch to the denoted state (the parent state
   * by default) and undo its proof.
   * @param state the state to which the command is applied
   * @return null
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Proof proof = state.getProof();
    if (label == null)
    {
      state = proof.getCurrent().getParent();
      if (state == null)
      {
        error("command not allowed in root state");
        return null;
      }
    }
    else
    {
      state = proof.getState(label);
      if (state == null)
      {
        error("there is no state with label [" + label + "]");
        return null;
      }
    }
    if (!state.isClosed())
    {
      error("cannot undo open state " + label + 
        " (use \"goto " + label + "\")");
      return null;
    }
    proof.cut(state);
    return null;
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (label != null)
    {
      out.print(" ");
      out.print(label);
    }
  }
}
