// ----------------------------------------------------------------------------
// TextInputStream.java
// Mapping of a Text widget to an InputStream object.
// $Id: TextInputStream.java,v 1.8 2009/10/18 16:46:39 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.SWT;

import java.io.*;
import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.events.*;

import fmrisc.ProofNavigator.Main;

import java.nio.charset.Charset;

/**
 * 
 * Mapping of a Text widget to an InputStream object.
 *
 */
public final class TextInputStream extends InputStream implements KeyListener
{
  // the widget
  private Text input;
  
  // the writer on which to echo the input (null, if no echo)
  private PrintWriter echo;
  
  // the buffer to hold the bytes not yet delivered by read()
  private byte[] buffer = new byte[0];
  
  // the input history for scrolling with cursor up / cursor down key
  private final int LEN = 20;  
  String[] history = new String[LEN];
  
  // the history position information
  private int number = 0;  // how many elements
  private int next = 0;    // where to add next element
  private int cursor = -1; // current cursor position (-1 if none)
  
  // the encoding character set
  private Charset charSet;
  
  /****************************************************************************
   * Create an input stream from a Text widget.
   * @param input the widget that provides the data for the input stream.
   * @param echo writer to which the input is to be echoed (null, if no echo)
   * @param charSet the character set to be used for encoding
   ***************************************************************************/
  public TextInputStream(Text input, PrintWriter echo, Charset charSet)
  {
    this.input = input;
    this.echo = echo;
    this.charSet = charSet;
    input.addKeyListener(this);
  }
  
  /****************************************************************************
   * Listener method that is invoked when key is released in widget.
   * @param e the event raised
   ***************************************************************************/
  public void keyReleased(KeyEvent e)
  {
  }
  
  /****************************************************************************
   * Listener method that is invoked when input is entered in widget.
   * @param e the event raised
   ***************************************************************************/
  public void keyPressed(KeyEvent e)
  {
    switch (e.keyCode)
    {
      // CTRL-a: goto beginning of line
      // CTRL-e: goto end of line
      // CTRL-k: kill everything from the caret position to end of line
      case 'a': case 'e': case 'k':
      {
        if ((e.stateMask & SWT.CTRL) == 0) return;
        e.doit = false;
        if (e.keyCode == 'a')
          input.setSelection(0,0);
        else if (e.keyCode == 'e')
        {
          int n = input.getText().length();
          input.setSelection(n,n);
        }
        else if (e.keyCode == 'k')
        {
          int p = input.getCaretPosition();
          input.setText(input.getText().substring(0, p));
          input.setSelection(p, p);
        }
        break;
      }
      // RETURN: send input to stream and save it in history
      case '\r': 
      {
        e.doit = false;
        history[next] = input.getText();
        next = (next+1) % LEN;
        if (number < LEN) number = number+1;
        cursor = -1;
        String text = input.getText()+input.getLineDelimiter();
        append(text);
        echo.print(text);
        echo.flush();
        input.setText("");
        break;
      }
      // ARROW_DOWN: cycle down in history buffer
      case SWT.ARROW_DOWN:
      {
        e.doit = false;
        if (number == 0) return;
        if (cursor == -1)
        {
          if (number < LEN)
            cursor = 0;
          else
            cursor = next;
        }
        else
        {
          cursor = cursor+1;
          if (cursor == number) cursor = 0;
        }
        input.setText(history[cursor]);
        break;
      }
      // ARROW_UP: cycle up in history buffer
      case SWT.ARROW_UP:
      {
        e.doit = false;
        if (number == 0) return;
        if (cursor == -1)
        {
          if (next == 0) 
            cursor = LEN-1;
          else
            cursor = next-1;
        }
        else
        {
          cursor = cursor-1;
          if (cursor == -1) cursor = number-1;
        }
        input.setText(history[cursor]);
        break;
      }
    }
  }
  
  /***************************************************************************
   * Append text to buffer and notify waiting reader.
   * @param text text to be appended to buffer.
   **************************************************************************/
  public synchronized void append(String text)
  {
    byte bytes[] = charSet.encode(text).array();
    int n = buffer.length;
    int m = bytes.length;
    byte[] buffer0 = new byte[n+m];
    System.arraycopy(buffer, 0, buffer0, 0, n);
    System.arraycopy(bytes, 0, buffer0, n, m);
    buffer = buffer0;
    notify();
  }
 
  /***************************************************************************
   * Get number of bytes available on stream.
   * @return number of bytes available on stream
   **************************************************************************/
  public synchronized int available()
  {
    if (buffer == null) return 0;
    return buffer.length;
  }

  /****************************************************************************
   * wait until text is availabe.
   ***************************************************************************/
  private synchronized void getText()
  {
    while (buffer != null && buffer.length == 0)
    {
      try
      {
        wait();
      }
      catch(InterruptedException e)
      {
      }
    }
  }

  /****************************************************************************
   * Write into buffer b at position offset at most length bytes.
   * @param b the byte buffer to be filled (may be null)
   * @param offset the distance from the start of the buffer where
   *        the filling shall start.
   * @param length the maximum number of bytes to be written to the buffer.
   * @return the actual number of bytes written.
   ***************************************************************************/
  public synchronized int read(byte b[], int offset, int length) 
  {
    getText();
    if (buffer == null) return 0;
    int n = buffer.length;
    if (n == 0) return -1;
    if (n > length) n = length;
    System.arraycopy(buffer, 0, b, offset, n);
    int length0 = buffer.length-n;
    byte[] buffer0 = new byte[length0];
    System.arraycopy(buffer, n, buffer0, 0, length0);
    buffer = buffer0;
    return n;
  }

  /***************************************************************************
   * Read a character from this stream.
   * @return the character
   **************************************************************************/
  public synchronized int read()
  {
    System.out.println("read() not supported");
    Main.exit(-1);
    return 0;
  }
  
  /****************************************************************************
   * close stream
   ***************************************************************************/
  public void close() throws IOException
  {
    buffer = null;
    synchronized(this) { notify(); }
  }
}
