// ----------------------------------------------------------------------------
// CVCLFormulaMask.java
// Mask certain formulas for CVCL as atomic.
// $Id: CVCLFormulaMask.java,v 1.7 2006/06/23 21:30:00 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving.CVCL;

import fmrisc.ProofNavigator.*;
import fmrisc.Syntax.*;
import fmrisc.Semantics.*;
import java.util.*;

public final class CVCLFormulaMask
{
  // reserved identifier starts with "_" to prevent collision with user ident
  private static final String MASKIDENTIFIER = "_MASK";
  
  // map of (formula text, mask identifier) pairs
  private HashMap repMap = new HashMap();
  
  // the value table holding all the generated symbols 
  private ValueTable valueTable = new ValueTable();
  
  // for generating unique masks
  private int counter = 0;
  
  // for collecting generated symbols
  private Vector symbols = new Vector();
  
  /***************************************************************************
   * Create formula mask with initial context.
   **************************************************************************/
  public CVCLFormulaMask()
  {
  }
     
  /***************************************************************************
   * Mask a formula.
   * @param f a formula expression
   * @return an expression that replaces the formula
   *         (either a reference or a reference application)
   **************************************************************************/
  public Expression getMask(Expression f)
  {   
    // construct lambda abstraction for free variables and its type
    Expression exp = f;
    Type type = Checking.getBooleanType();
    Vector freeVars = FreeVariables.compute(exp);
    int n = freeVars.size();
    if (n > 0)
    {
      Type[] atypes = new Type[n];
      TypedIdentifier[] params = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        ValueSymbol symbol = (ValueSymbol)(freeVars.elementAt(i));
        ValueDeclIdentifier var = symbol.getValueDeclIdentifier();
        atypes[i] = symbol.getType();
        params[i] = Construct.typedIdentifier(var, atypes[i]);
      }
      type = Checking.canonical(Construct.functionType(atypes, type));
      exp = Construct.lambdaTerm(params, f);
    }
    
    // prevent recursion in determining canonical text
    State.setFormulaMask(null);
    String text = ASTUtil.canonicalText(exp);
    State.setFormulaMask(this);
    
    // determine mask identifier
    ValueDeclIdentifier ident = (ValueDeclIdentifier)repMap.get(text);
    boolean found = ident != null;
    if (!found) 
    {
      ident = Construct.valueDeclIdentifier(MASKIDENTIFIER + counter);
      counter = counter+1;
      repMap.put(text, ident);
      ValueSymbol symbol = valueTable.put(ident, type, exp.instantiate(), 0);
      symbols.add(symbol);
    }
    
    // construct mask expression
    Expression mask = Construct.reference(ident);
    if (n > 0)
    {
      Expression[] args = new Expression[n];
      for (int i=0; i<n; i++)
      {
        ValueSymbol symbol = (ValueSymbol)(freeVars.elementAt(i));
        ValueDeclIdentifier var = symbol.getValueDeclIdentifier();
        args[i] = Construct.reference(var);
      }
      mask = Construct.applicationExpression(mask, args);
    }
    return mask;
  }
  
  /****************************************************************************
   * Get formula associated to mask identifier
   * @param mask identifier of a mask
   * @return formula expression to which mask is bound (null, if none)
   ***************************************************************************/
  public Expression getFormula(Identifier mask)
  {
    ValueSymbol symbol = valueTable.getSymbol(mask);
    if (symbol == null) return null;
    return symbol.getValue();
  }
  
  /***************************************************************************
   * Reset collection of generated mask symbols.
   **************************************************************************/
  public void resetSymbols()
  {
    symbols = new Vector();
  }
  
  /***************************************************************************
   * Return mask symbols generated since last call of resetSymbols().
   * @return collection of ValueSymbol objects representing the mask symbols.
   **************************************************************************/
  public Collection getSymbols()
  {
    return symbols;
  }
}
