// ----------------------------------------------------------------------------
// ProofState.java
// A proof state.
// $Id: ProofState.java,v 1.38 2007/10/23 20:10:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving;

import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.Communication.*;
import java.util.*;
import java.io.*;

/**
 * The state of a proof.
 * A proof state E(AX): A1,...,An |- B1,...Bm consists of 
 * - an environment E of function/predicate cohnstant definitions
 * - a sequence of axioms assumed true in that state
 * - a sequence of formulas ("assumptions") A1,...,An
 * - a sequence of formulas (potential "goals") B1,...,Bm.
 * - potentially, a sequence of formulas (the counterexample) C1,...,Co.
 * The state logically represents the formula 
 *   AX => (~A1 \/ ... \/ ~An) \/ (B1 \/ ... \/ Bm)
 * i.e. 
 *   AX => (A1 /\ ... /\ An) => (B1 \/ ... \/ Bm).
 * If given, the counterexample represents a formula
 *   C1 /\ ... /\ Co
 * which implies AX /\ (A1 /\ ... /\ An) /\ (~B1 /\ ... /\ ... /\ ~Bn).
 */
public final class ProofState
{
  // proof to which the state belongs
  private Proof proof;
  
  // path that uniquely identifies the state
  private String path;

  // label that hashes path to a short word
  private String label;
  
  // the parent state
  private ProofState parent;
  
  // the environment
  private Environment env;
  
  // the collection of formulas in the state
  private Formula[] axioms = {};
  private Formula[] assumptions = {};
  private Formula[] goals = {};
  
  // sequence of expressions representing conjunction of formulas
  private Expression[] counterExample = null;
  
  // the command applied to yield the children states
  private Command command;
  
  // the command to be executed if proof is redone
  private Command redoCommand;
  
  // the children states
  private ProofState[] children;
 
  // length of proof state label in number of characters
  private static final int len=3; 
  
  // counter to make state labels unique
  private static int counter=0;   
  
  // unique name table for skolem constants
  private UniqueNameTable skolemTable;
  
  // listener that is notified when state has changed
  private ProofStateListener listener;
  
  // automatic simplification in proof state on or off
  private boolean autoSimplify;
  
  /****************************************************************************
   * construct a root state from a single goal and without assumptions
   * @param proof the proof to which the state belongs
   * @param env the environment in which all constants are declared
   * @param goal the goal formula
   * @param autoSimplify true iff automatic simplification is on in state
   ***************************************************************************/
  public ProofState(Proof proof, Environment env, Expression goal,
    boolean autoSimplify)
  {
    this.proof = proof;
    this.path = proof.getSymbol().getIdentifier().toString();
    this.label = hash(this.path);
    this.env = env;
    this.autoSimplify = autoSimplify;
    
    // type checking conditions as state axioms
    Expression[] tccs = env.getTCCs();
    if (tccs != null)
    {
      int n = tccs.length;
      Formula[] formulas = new Formula[n];
      for (int i=0; i<n; i++)
        formulas[i] = new Formula(this, tccs[i], false);
      this.axioms = formulas;
    }
    
    // user-declared axioms as assumptions
    Vector v = new Vector();
    Iterator iterator = env.getFormulaSymbols().iterator();
    while (iterator.hasNext())
    {
      FormulaSymbol symbol0 = (FormulaSymbol)(iterator.next());
      if (symbol0.isAxiom())
        v.add(new Formula(this, symbol0.getFormula(), false));
    }
    this.assumptions = Formula.toArray(v);
    
    // formula to be proved as goal
    this.goals = new Formula[]{ new Formula(this, goal, true) };
    
    // skolemization table
    skolemTable = new UniqueNameTable(env);
  }

  /****************************************************************************
   * construct a proof state without path and formulas yet
   * @param parent the parent state
   * @param env the environment in which all constants are declared
   ***************************************************************************/
  public ProofState(ProofState parent, Environment env)
  {
    this.proof = parent.getProof();
    this.parent = parent;
    this.env = env;
    skolemTable = new UniqueNameTable(env);
    this.axioms = parent.axioms;
    this.autoSimplify = parent.autoSimplify;
  }
  
  /****************************************************************************
   * Construct a skeleton proof state
   * (all fields but the proof remain uninitialized).
   * @param proof the proof to which the state shall belong.
   ***************************************************************************/
  public ProofState(Proof proof)
  {
    this.proof = proof;
  }
  
  /***************************************************************************
   * return label of proof state
   * @return the proof state label
   **************************************************************************/
  public String getLabel()
  {
    return label;
  }
  
  /***************************************************************************
   * Hash state path to a short label.
   * @param a state path
   * @return a short word
   **************************************************************************/
  private String hash(String path)
  {
    String label = ProofUtil.hash(path, len);
    if (proof.getState(label) != null)
    {
      // resolve the unlikely case of a hash collision
      label = label + "-" + counter;
      counter = counter+1;
    }
    return label;
  }
  
  /****************************************************************************
   * get proof to which proof state belong
   * @return the proof
   ***************************************************************************/
  public Proof getProof()
  {
    return proof;
  }
  
  /****************************************************************************
   * get path of proof state
   * @return the path of the state
   ***************************************************************************/
  public String getPath()
  {
    return path;
  }
  
  /****************************************************************************
   * set path of proof state
   * @param path the path of the state (must be unique in the current proof)
   ***************************************************************************/
  public void setPath(String path)
  {
    this.path = path;
    this.label = hash(this.path);
  }
  
  /****************************************************************************
   * get parent of proof state
   * @return the parent of the state
   ***************************************************************************/
  public ProofState getParent()
  {
    return parent;
  }
  
  /****************************************************************************
   * set parent of proof state
   * @param parent the parent of the state
   ***************************************************************************/
  public void setParent(ProofState parent)
  {
    this.parent = parent;
  }
  /****************************************************************************
   * get environment of proof state
   * @return the environment
   ***************************************************************************/
  public Environment getEnvironment()
  {
    return env;
  }
  
  /****************************************************************************
   * set environment of proof state
   * @param env the environment.
   ***************************************************************************/
  public void setEnvironment(Environment env)
  {
    this.env = env;
  }
  
  /****************************************************************************
   * get axioms of proof state
   * @return the axioms of the proof state
   ***************************************************************************/
  public Formula[] getAxioms()
  {
    return axioms;
  }
  
  /****************************************************************************
   * Set axioms of proof state.
   * @param formulas the axioms of the state
   ***************************************************************************/
  public void setAxioms(Formula[] formulas)
  {
    this.axioms = formulas;
  }
  
  /****************************************************************************
   * get assumptions of proof state
   * @return the assumptions of the proof state
   ***************************************************************************/
  public Formula[] getAssumptions()
  {
    return assumptions;
  }
  
  /****************************************************************************
   * set assumptions of proof state
   * @param formulas the assumptions of the state
   ***************************************************************************/
  public void setAssumptions(Formula[] formulas)
  {
    this.assumptions = formulas;
  }
  
  /****************************************************************************
   * get goals of proof state
   * @return the goals of the proof state
   ***************************************************************************/
  public Formula[] getGoals()
  {
    return goals;
  }

  /****************************************************************************
   * set goals of proof state
   * @param formulas the goals of the state
   ***************************************************************************/
  public void setGoals(Formula[] formulas)
  {
    this.goals = formulas;
  }
  
  /****************************************************************************
   * Get counterexample for state.
   * @return the counterexample of the state (possibly null).
   ***************************************************************************/
  public Expression[] getCounterExample()
  {
    return counterExample;
  }

  /****************************************************************************
   * Set counterexample for state.
   * @param formulas sequence of formulas interpreted as a conjunction.
   ***************************************************************************/
  public void setCounterExample(Expression[] formulas)
  {
    this.counterExample = formulas;
  }
  
  /****************************************************************************
   * get command of proof state
   * @return command that generated the children states (null, if no children)
   ***************************************************************************/
  public Command getCommand()
  {
    return command;
  }
  
  /****************************************************************************
   * set command of proof state
   * @param command the command applied to the proof state
   ***************************************************************************/
  public void setCommand(Command command)
  {
    this.command = command;
  }
  
  /****************************************************************************
   * get automatic simplification status of proof state
   * @return true iff automatic simplification is switched on
   ***************************************************************************/
  public boolean useAutoSimplify()
  {
    return autoSimplify;
  }
  
  /****************************************************************************
   * Set automatic simplification status of proof state
   * @param on true iff automatic simplification is to be switched on
   ***************************************************************************/
  public void setAutoSimplify(boolean on)
  {
    autoSimplify = on;
  }
  
  /****************************************************************************
   * get children of proof state
   * @return the children of the current state (null, if none)
   ***************************************************************************/
  public ProofState[] getChildren()
  {
    if (command == null) return null;
    return children;
  }
  
  /****************************************************************************
   * set children of proof state
   * @param command the command that generated the children
   * @param children the new children of the state
   ***************************************************************************/
  public void setChildren(Command command, ProofState[] children)
  {
    this.command = command;
    this.redoCommand = null;
    this.children = children;
  }
  
  /***************************************************************************
   * returns formula (assumption or goal) with denoted label from sequence
   * @param seq sequence of formulas
   * @param label a formula label
   * @return formula from seq with that label or null, if there is none
   **************************************************************************/
  private static Formula getFormula(Formula[] formulas, String label)
  {
    int n = formulas.length;
    for (int i=0; i<n; i++)
    {
      Formula f = formulas[i];
      if (label.equals(f.getLabel())) return f;
    }
    return null;
  }
  
  /***************************************************************************
   * returns assumption with denoted label
   * @return formula or null, if there is no assumption with that label
   **************************************************************************/
  public Formula getAssumption(String label)
  {
    return getFormula(assumptions, label);
  }
  
  /***************************************************************************
   * returns goal with denoted label
   * @return formula or null, if there is no goal with that label
   **************************************************************************/
  public Formula getGoal(String label)
  {
    return getFormula(goals, label);
  }
  
  /***************************************************************************
   * returns formula (assumption or goal) with denoted label
   * @return formula or null, if there is no formula with that label)
   **************************************************************************/
  public Formula getFormula(String label)
  {
    Formula f = getAssumption(label);
    if (f != null) return f;
    return getGoal(label);
  }
  
  /***************************************************************************
   * Get unique name table to be used for creating skolemization constants.
   * @return the unique name table
   *************************************************************************/
  public UniqueNameTable getSkolemTable()
  {
    return skolemTable;
  }
  
  /****************************************************************************
   * Set listener that is notified when proof state has changed.
   * @param listener the listener that is notified.
   ***************************************************************************/
  public void setListener(ProofStateListener listener)
  {
    this.listener = listener;
  }
  
  /****************************************************************************
   * Get listener that is notified when proof state has changed.
   * @return the listener that is notified.
   ***************************************************************************/
  public ProofStateListener getListener()
  {
    return listener;
  }
  
  /****************************************************************************
   * Signal that state has changed updating presentation of this state.
   * @param file true iff also presentation on file is to bechanged.
   ***************************************************************************/
  public void hasChanged(boolean file)
  {
    if (listener != null) listener.hasChanged();
    // remove here, if performed in ProofTreeItem.addChildren()
    if (file) State.getPresenter().writeProofState(this);
  }
  
  /***************************************************************************
   * print proof state on output stream 
   * @param out the stream on which to print the formula
   **************************************************************************/
  public void print(PrintWriter out)
  {
    String auto;
    if (autoSimplify)
      auto = "(autosimp)";
    else
      auto = "(no autosimp)";
    if (command == null)
      out.println("Proof state [" + label + "] " + auto);
    else
      out.println("Proof state [" + label + "] " + auto + 
        ": " + command.toString());
    Collection symbols = env.getValueSymbols();
    if (!symbols.isEmpty())
    {
      out.print("Constants: ");
      Iterator iterator = symbols.iterator();
      ValueSymbol symbol = (ValueSymbol)iterator.next();
      symbol.getIdentifier().print(out);
      out.print(": ");
      symbol.getType().print(out);
      while (iterator.hasNext())
      {
        out.print(", ");
        symbol = (ValueSymbol)iterator.next();
        symbol.getIdentifier().print(out);
        out.print(": ");
        symbol.getType().print(out);
      }
      out.println(".");
    }
    int l = axioms.length;
    if (l != 0)
    {
      out.print("Axioms: ");
      for (int i=0; i<l; i++)
      {
        axioms[i].getExpression().print(out);
        if (i+1 < l) out.print(", ");
      }
      out.println(".");
    }
    
    int n = assumptions.length;
    for (int i=0; i<n; i++)
    {
      Formula f = assumptions[i];
      f.print(out);
      out.println();
    }
    out.println("-------------------------------------------------------");
    int m = goals.length;
    for (int i=0; i<m; i++)
    {
      Formula f = goals[i];
      f.print(out);
      out.println();
    }
    
    if (counterExample != null)
    {
      out.println("This may be a counterexample:");
      ASTUtil.toConjunction(counterExample).print(out);
      out.println();
    }
    Main.setProofState(this);
  }
 
  /***************************************************************************
   * close the state.
   * @param reason explanation why state is closed.
   **************************************************************************/
  public void close(String reason)
  {
    command = new Proved(reason);
    children = null;
  }
  
  /***************************************************************************
   * open the state.
   **************************************************************************/
  public void open()
  {
    redoCommand = command;
    command = null;
    // reset name table such that same names will be generated in future
    skolemTable = new UniqueNameTable(env);
    // leave children unchanged for redoing command
    // children = null;
    hasChanged(true);
  }
  
  /***************************************************************************
   * Redo the proof at the denoted state (if possible)
   **************************************************************************/
  public void redo()
  {
    if (redoCommand == null) return;
    command = redoCommand;
    redoCommand = null;
    hasChanged(true);
  }
  
  /***************************************************************************
   * Label the children of the current state.
   **************************************************************************/
  public void labelChildren()
  {
    ProofState[] c = getChildren();
    if (c == null) return;
    Presenter presenter = State.getPresenter();
    String base = getPath() + ".";
    int n = c.length;
    for (int i=0; i<n; i++)
    {
      String path = base + i;
      ProofState cstate = c[i];
      cstate.setPath(path);
      presenter.writeProofState(cstate);
    }
  }
  
  /***************************************************************************
   * command extends current state by adding and labeling children
   * @param command the command extending the state
   * @param children the children of the proof state 
   **************************************************************************/
  public void extend(Command command, ProofState[] children)
  {
    setChildren(command, children); 
    labelChildren();
    proof.extend(this, children);
    hasChanged(true);
  }
  
  /***************************************************************************
   * check whether proof state is closed 
   * @return true iff proof state is closed
   *************************************************************************/
  public boolean isClosed()
  {
    return command != null;
  }
  
  /***************************************************************************
   * simplify the current proof state and potentially close it.
   * @param hard true iff hard (and time-consuming) simplification is tried.
   * @return true iff everything went okay.
   **************************************************************************/
  public boolean simplify(boolean hard)
  {
    if (isClosed()) return true;
    
    boolean okay = ProverReasoning.simplify(this, hard, hard);
    if (!okay) return false;
    if (isClosed()) return true;
    
    // labels of axioms, assumptions and goals, new axioms, assumptions and goals
    HashSet xlabels = new HashSet();
    HashSet alabels = new HashSet();
    HashSet glabels = new HashSet();
    Vector axioms0 = new Vector();
    Vector assumptions0 = new Vector();
    Vector goals0 = new Vector();

    // remove duplicate axioms, deal with axioms TRUE and FALSE
    boolean changed2 = false;
    int o = axioms.length;
    for (int i=0; i<o; i++)
    {
      Formula formula = axioms[i];
      Expression exp = formula.getExpression();
      if (!(exp instanceof Logical))
      {
        String label = formula.getLabel();
        // duplicate axiom is removed
        if (xlabels.contains(label))
          changed2 = true;
        else
        {
          xlabels.add(label);
          axioms0.add(formula);
        }
        continue;
      }
      changed2 = true;
      // TRUE is discarded, FALSE means proved
      if (!((Logical)exp).getValue())
      {
        close("Axiom " + formula.getLabel() + " is FALSE");
        return true;
      }
    }
    if (changed2) setAxioms(Formula.toArray(axioms0));
    
    // remove duplicate assumptions, deal with assumptions TRUE and FALSE
    boolean changed0 = false;
    int n = assumptions.length;
    for (int i=0; i<n; i++)
    {
      Formula formula = assumptions[i];
      Expression exp = formula.getExpression();
      if (!(exp instanceof Logical))
      {
        String label = formula.getLabel();
        // duplicate assumption is removed
        if (alabels.contains(label))
          changed0 = true;
        else
        {
          alabels.add(label);
          assumptions0.add(formula);
        }
        continue;
      }
      changed0 = true;
      // TRUE is discarded, FALSE means proved
      if (!((Logical)exp).getValue())
      {
        close("Assumption " + formula.getLabel() + " is FALSE");
        return true;
      }
    }
    if (changed0) setAssumptions(Formula.toArray(assumptions0));
    
    // some goal is TRUE or FALSE or occurs among the axioms/assumptions?
    boolean changed1 = false;
    int m = goals.length;
    for (int i=0; i<m; i++)
    {
      Formula formula = goals[i];
      Expression exp = formula.getExpression();
      if (!(exp instanceof Logical))
      {
        String label = formula.getLabel();
        if (xlabels.contains(label))
        {
          // goal occurs among axioms
          close("Goal " + formula.getLabel() + " equals some axiom");
          return true;
        }
        if (alabels.contains(label))
        {
          // goal occurs among assumptions
          close("Goal " + formula.getLabel() + " equals some assumption");
          return true;
        }
        // duplicate formulas are removed
        if (alabels.contains(label) || glabels.contains(label))
          changed1 = true;
        else
        {
          glabels.add(label);
          goals0.add(formula);
        }
        continue;
      }
      changed1 = true;
      // TRUE means proved, FALSE is discarded
      if (((Logical)exp).getValue())
      {
        close("Goal " + formula.getLabel() + " is TRUE");
        return true;
      }
    }
    if (changed1) setGoals(Formula.toArray(goals0));
    return true;
  }
  
  /***************************************************************************
   * Copy a sequence of formulas in this state.
   * @param formulas a sequence of formulas
   * @return a sequence of formulas all of which belong to this state such that
   * this sequence equals formulas.
   **************************************************************************/
  public Formula[] copy(Formula[] formulas)
  {
    int n = formulas.length;
    Formula[] formulas0 = new Formula[n];
    for (int i=0; i<n; i++)
      formulas0[i] = new Formula(this, formulas[i]);
    return formulas0;
  }
  
  /***************************************************************************
   * Adds a formula to a sequence of formulas in this state.
   * @param formulas a sequence of formulas 
   * @param formula a formula belonging to this state
   * @return a sequence of formulas all of which belong to this state such that
   * this sequence equals formulas appended by formula
   **************************************************************************/
  public Formula[] add(Formula[] formulas, Formula formula)
  {
    int n = formulas.length;
    Formula[] formulas0 = new Formula[n+1];
    for (int i=0; i<n; i++)
      formulas0[i] = new Formula(this, formulas[i]);
    formulas0[n] = formula;
    return formulas0;
  }
  
  /***************************************************************************
   * Adds a sequence of formulas to a sequence of formulas in this state.
   * @param formulas a sequence of formulas 
   * @param formulas0 a sequence of formulas belonging to this state
   * @return a sequence of formulas all of which belong to this state such that
   * this sequence equals formulas appended by formulas0
   **************************************************************************/
  public Formula[] add(Formula[] formulas, Formula[] formulas0)
  {
    int n = formulas.length;
    int n0 = formulas0.length;
    Formula[] formulas1 = new Formula[n+n0];
    for (int i=0; i<n; i++)
      formulas1[i] = new Formula(this, formulas[i]);
    for (int i=0; i<n0; i++)
      formulas1[n+i] = formulas0[i];
    return formulas1;
  }
  
  /***************************************************************************
   * Remove a formula from a sequence of formulas in this state.
   * @param formulas a sequence of formulas
   * @param formula a formula occuring in that sequence once
   * @return a sequence of formulas all of which belong to this state such that
   * this sequence equals formulas without formula
   **************************************************************************/
  public Formula[] remove(Formula[] formulas, Formula formula)
  {
    int n = formulas.length;
    Formula[] formulas0 = new Formula[n-1];
    int p=0;
    for (int i=0; i<n; i++)
    {
      if (formulas[i] == formula) continue;
      formulas0[p] = new Formula(this, formulas[i]);
      p = p+1;
    }
    return formulas0;
  }
  
  /***************************************************************************
   * Replace a formula in a sequence of formulas in this state.
   * @param formulas a sequence of formulas
   * @param formula a formula occuring in that sequence once
   * @param rformula a formula belonging to this state
   * @return a sequence of formulas all of which belong to this state such that
   * this sequence equals formulas without formula but includes rformula
   **************************************************************************/
  public Formula[] replace(Formula[] formulas, Formula formula, Formula rformula)
  {
    int n = formulas.length;
    Formula[] formulas0 = new Formula[n];
    for (int i=0; i<n; i++)
    {
      if (formulas[i] == formula) 
        formulas0[i] = rformula;
      else
        formulas0[i] = new Formula(this, formulas[i]);
    }
    return formulas0;
  }
  
  /***************************************************************************
   * Print proof state including children.
   * @param out the stream to print the state to.
   * @param level the level of the proof state (0 = root).
   **************************************************************************/
  public void print(PrintWriter out, int level)
  {
    for (int i=0; i<level; i++)
      out.print("  ");
    out.print("[" + label + "]");
    // out.print(path);
    if (this == proof.getCurrent())
      out.print("***");
    else if (!isClosed()) 
      out.print("*");
    else
      out.print(": " + command.toString());
    out.println();
    ProofState[] c = getChildren();
    if (c == null) return;
    int n = c.length;
    for (int i=0; i<n; i++)
      c[i].print(out, level+1);
  }
  
  /***************************************************************************
   * Add open state and all open children recursively to vector.
   * @param vector vector receiving all open proof states.
   **************************************************************************/
  public void addOpen(Vector vector)
  {
    if (!isClosed()) vector.add(this);
    ProofState[] c = getChildren();
    if (c == null) return;
    int n = c.length;
    for (int i=0; i<n; i++)
      c[i].addOpen(vector);
  }
}
