// ----------------------------------------------------------------------------
// BinaryExpression.java
// Abstract base class of binary expressions
// $Id: BinaryExpression.java,v 1.12 2006/02/28 14:39:04 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;
import java.io.*;

/**
 * Abstract base class of binary expressions
 */
public abstract class BinaryExpression extends ExpressionBase
{
  private String op;         // name of operation
  private Expression first;  // first subexpression
  private Expression second; // second subexpression

  /****************************************************************************
   * constructs binary expression from op, first, and second
   * @param op the name of the operation used for printing
   * @param first the first operand
   * @param second the second operand
   ***************************************************************************/
  public BinaryExpression(String op, Expression first, Expression second)
  {
    this.op = op;
    this.first = first;
    this.second = second;
  }
  
  /***************************************************************************
   * Returns first subexpression
   * @return the first subexpression
   **************************************************************************/
  public Expression getFirst()
  {
    return first;
  }
  
  /***************************************************************************
   * Returns second subexpression
   * @return the second subexpression
   **************************************************************************/
  public Expression getSecond()
  {
    return second;
  }
  
  /***************************************************************************
   * Prints text representation of tree on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    first.printPriority(getPriority(), out);
    out.print(op);
    second.printPriority(getPriority(), out);
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    return construct(getFirst().instantiate(), getSecond().instantiate());
  }
  
  /***************************************************************************
   * Construct binary expression of the same kind as this expression.
   * @param first an expression.
   * @param second an expression.
   * @return a binary expression of the same kind as this expression
   *         with subexpressions first and second.
   **************************************************************************/
  public abstract BinaryExpression construct(
    Expression first, Expression second);
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST first0 = first.accept(visitor);
    AST second0 = second.accept(visitor);
    if (!visitor.isCloned()) return null;
    return construct((Expression)first0, (Expression)second0);
  }
}