// ----------------------------------------------------------------------------
// LetType.java
// Handling of types with local bindings.
// $Id: LetType.java,v 1.4 2006/02/23 19:51:05 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import java.io.*;

/**
 * Handling of terms with local bindings.
 */
public final class LetType extends TypeBase
{
  private TypeDeclaration[] declarations; // the local bindings
  private Type base;                      // the base type

  /****************************************************************************
   * Construct let types of declarations and base types.
   * @param declarations the local bindings in the types.
   * @param base the base type.
   ***************************************************************************/
  public LetType(TypeDeclaration[] declarations, Type base)
  {
    this.declarations = declarations;
    this.base = base;
  }
  
  /****************************************************************************
   * get declarations
   * @return the declaration sequence
   ***************************************************************************/
  public TypeDeclaration[] getDeclarations()
  {
    return declarations;
  }
  
  /****************************************************************************
   * get base type
   * @return the base type
   ***************************************************************************/
  public Type getBase()
  {
    return base;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Type instantiateType()
  {
    TypeDeclaration[] decls = getDeclarations();
    int n = decls.length;
    TypeDeclaration[] decls0 = new TypeDeclaration[n];
    for (int i=0; i<n; i++)
    {
      decls0[i] = decls[i].instantiate();
    }
    Type base0 = getBase().instantiateType();
    return Construct.letType(decls0, base0);
  }
  
  /***************************************************************************
   * Prints text representation of tree on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print("LET ");
    int n = declarations.length;
    for (int i=0; i<n-1; i++)
    {
      declarations[i].print(out);
      out.print(", ");
    }
    declarations[n-1].print(out);
    out.print(" IN ");
    base.print(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null)
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = declarations.length;
    TypeDeclaration[] declarations0 = new TypeDeclaration[n];
    for (int i=0; i<n; i++)
      declarations0[i] = (TypeDeclaration)declarations[i].accept(visitor);
    AST base0 = base.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.letType(declarations0, (Type)base0);
  }
}
